/*
*  Sparkline Visual
*
*  Copyright (c) Prologika, LLC
*  https://prologika.com
*  All rights reserved. 
*  MIT License
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the ""Software""), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*   
*  The above copyright notice and this permission notice shall be included in 
*  all copies or substantial portions of the Software.
*   
*  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
*  THE SOFTWARE.

Description
A sparkline is a small embedded line graph for showing a trend usually over time. Configure the sparkline go group data by a 
single category field, such as CalendarYear, and a single measure, such as SalesAmount. You can configure the line color, width, and even animate the graph.

To use the sparkline:
1. Import the sparkline.pbiviz file in Power BI or Power BI Desktop to add it to the report from the Visualizations pane.
2. Add a single grouping field in the Category area and a single measure in the Value area. The sparkline should show a graph.
3. You can change formatting options in the Format of the Visualizations pane, including the line color, width, 
   and animation options if you want the graph to animate.

The sparkline implementation is fully described in the book "Applied Power BI" by Teo Lachev.
*/
"use strict";

import "core-js/stable";
import "./../style/visual.less";
import powerbi from "powerbi-visuals-api";
import VisualConstructorOptions = powerbi.extensibility.visual.VisualConstructorOptions;
import VisualUpdateOptions = powerbi.extensibility.visual.VisualUpdateOptions;
import IVisual = powerbi.extensibility.visual.IVisual;
import EnumerateVisualObjectInstancesOptions = powerbi.EnumerateVisualObjectInstancesOptions;
import VisualObjectInstanceEnumeration = powerbi.VisualObjectInstanceEnumeration;
import { VisualSettings } from "./settings";
import IVisualHost = powerbi.extensibility.visual.IVisualHost;
import DataView = powerbi.DataView;
import SelectionManager = powerbi.extensibility.ISelectionManager;

import * as d3 from "d3";
import * as ts from './tooltipServiceWrapper'
//import * as ti from './tooltipInterfaces'

type Selection<T extends d3.BaseType> = d3.Selection<T, any,any, any>;

export interface SparklineModel {
    color: string;
    size: number;
    data: number[];
    animate: boolean;
    delay: number;
    duration: number;
    selector: powerbi.extensibility.ISelectionId;
 }

export class Visual implements IVisual {
    private host: IVisualHost;
    private timerId: any;
    private svg: Selection<SVGElement>;
    private root: Selection<HTMLElement>;
    private selectionManager: SelectionManager;
    private tooltipServiceWrapper: ts.ITooltipServiceWrapper;
    private settings: VisualSettings;

    public static converter(
        dataView: DataView,
        settings: VisualSettings
    ): SparklineModel {

        if (!dataView) {
            return;
        }

        if (!Visual.isDataViewValid(dataView)) {
            return;
        }

        var categoryValuesLen = 0;
        var dataViewCategorical = dataView.categorical;

        var categories = dataViewCategorical.categories[0]; // get to categories
        var values = dataViewCategorical.values && dataViewCategorical.values[0].values; // get to measures
        var categoryValues = categories && categories.values;
        var categoryValuesLen = categoryValues && categoryValues.length;
        var tooltipString: string = "";

        var viewModel: SparklineModel = {
            color: settings.general.fill,
            size: settings.general.size,
            data: values as number[],
            animate: settings.labels.show,
            duration: settings.labels.duration,
            delay: settings.labels.delay,
            selector: null            
        };
        return viewModel;
    } 

    // validates data view to ensure we have one group and one measure
    private static isDataViewValid(dataView: DataView): boolean {
        if (dataView && dataView.categorical && dataView.categorical.categories && dataView.metadata
            && dataView.categorical.categories[0] && dataView.categorical.values
        ) {
            return true;
        }
        return false;
    }
    
    constructor(options: VisualConstructorOptions) {
        this.selectionManager = options.host.createSelectionManager();
        this.root = d3.select(options.element);
        this.tooltipServiceWrapper = ts.createTooltipServiceWrapper(options.host.tooltipService, options.element);        

        this.svg = this.root
            .append('svg')
            .classed('sparkline', true)
            .attr('height', options.element.clientHeight)
            .attr('width', options.element.clientWidth);
    }

        // Update method where the visual draw actually happens
        public update(options: VisualUpdateOptions) {
                       
            this.svg.selectAll("path").remove(); // clear existing line
            if (!options.dataViews || !options.dataViews[0]) return;
            const dataView: DataView = options.dataViews[0];
            this.settings = VisualSettings.parse<VisualSettings>(dataView);
            var viewport = options.viewport;
            var viewModel: SparklineModel = Visual.converter(dataView, this.settings);

            if (!viewModel) return;
            if (viewport.height < 0 || viewport.width < 0) return;
            var graph = this.svg;

            // stop animation if graph is animating for update changes to take effect
            this.stopAnimation();
            // resize draw area to fit visualization frame
            this.svg.attr("width", viewport.width);
            this.svg.attr("height", viewport.height);

            var data = viewModel.data;
            // X scale fits values for all data elements; domain property will scale the graph width
            var x = d3.scaleLinear().domain([0, data.length - 1]).range([0, viewport.width]);
            // Y scale will fit values from min to max calibrated to the graph higth
            var y = d3.scaleLinear().domain([Math.min.apply(Math, data), Math.max.apply(Math, data)]).range([0, viewport.height]);

            // create a line
            var line = d3.line<number>()
                
                .curve(d3.curveBasis) // smooth line
                // assign the X function to plot on X axis
                .x(function (d, i) {
                    // enable the next line when debugging to output X coordinate
                    // console.log('Plotting X value for data point: ' + d + ' using index: ' + i + ' to be at: ' + x(i) + ' using xScale.');
                    return x(i);
                })
                .y(function (d) {
                    // enable the next line when debugging to output X coordinate
                    // console.log('Plotting Y value for data point: ' + d + ' to be at: ' + y(d) + " using yScale.");
                    return viewport.height - y(d); // values are plotted from the top so reverse the scale
                })
            
            // display the line by appending an svg:path element with the data line we created above
            var path = this.svg.append("svg:path")
                .attr("d", line(data))
                .attr('stroke-width', function (d) { return viewModel.size })
                .attr('stroke', function (d) { return viewModel.color })
                ;

            // the animation function
            function redrawWithAnimation() {
                var totalLength = (<SVGPathElement>path.node()).getTotalLength();
                graph.selectAll("path")
                    .data([data]) // set the new data
                    .attr("d", line)
                    .attr("stroke-dasharray", totalLength + " " + totalLength)
                    .attr("stroke-dashoffset", totalLength)
                    .transition()
                    .duration(viewModel.duration)
                    .ease(d3.easeLinear)
                    .attr("stroke-dashoffset", 0);
            }

            if (viewModel.animate) {
                this.timerId = window.setInterval(function () {
                    redrawWithAnimation();
                }, viewModel.delay);
            }
            
            this.tooltipServiceWrapper.addTooltip(this.svg.selectAll("path"),
                (tooltipEvent: ts.TooltipEventArgs<number>) => this.getTooltipData(tooltipEvent, dataView, viewport),
                (tooltipEvent: ts.TooltipEventArgs<number>) => null);
            
        }
        public destroy(): void {
            this.svg = null;
            this.root = null;
            this.timerId = null;
        }

        private stopAnimation() {
            if (this.timerId) {
                window.clearInterval(this.timerId);
                this.timerId = null;
            }
        }

    private static parseSettings(dataView: DataView): VisualSettings {
        return <VisualSettings>VisualSettings.parse(dataView);
    }

    /** 
     * This function gets called for each of the objects defined in the capabilities files and allows you to select which of the 
     * objects and properties you want to expose to the users in the property pane.
     * 
     */
    public enumerateObjectInstances(options: EnumerateVisualObjectInstancesOptions): VisualObjectInstanceEnumeration {
        const settings: VisualSettings = this.settings || <VisualSettings>VisualSettings.getDefault();
        return VisualSettings.enumerateObjectInstances(this.settings || VisualSettings.getDefault(), options);
     }
    
    private getTooltipData(value: any, dataView:DataView, viewport:powerbi.IViewport): powerbi.extensibility.VisualTooltipDataItem[] {
        var model = Visual.converter(dataView, this.settings);
        var tooltip = <ts.TooltipEventArgs<number>> value;
        var x = tooltip.coordinates[0];
        var width = viewport.width;
        var i = Math.floor((x/width) * model.data.length); // find the closest data point from coordinates
        var d = model.data[i];

        return [{
            displayName: dataView.categorical.categories[0].values[i].toString(),
            value: d.toString()
        }];
    } 
    
}